"""Tests the replay buffer module."""
from copy import deepcopy

import numpy as np
import pytest

from wingman import ReplayBuffer


def is_equivalent_tuple(item1, item2):
    """Checks whether a two tuples of np.ndarrays are equivalent to each other."""
    equivalence = True
    for i1, i2 in zip(item1, item2):
        equivalence = (i1 == i2).all() and equivalence
    return equivalence


@pytest.mark.parametrize("random_rollover", [True, False])
def test_bulk(random_rollover):
    """Tests repeatedly bulking the buffer and whether it rollovers correctly."""
    bulk_size = 7
    mem_size = 11
    element_shapes = [(3, 3), (3,), ()]
    memory = ReplayBuffer(mem_size=mem_size)

    for iteration in range(10):
        # try to stuff:
        # a) (bulk_size, 3) array
        # b) (bulk_size,) array
        data = []
        for shape in element_shapes:
            data.append(np.random.randn(bulk_size, *shape))
        memory.push(data, bulk=True, random_rollover=random_rollover)

        # reverse the data to make indexing for checking easier
        reversed_data = list(map(list, zip(*data)))

        # if random rollover and we're more than full, different matching technique
        if random_rollover and memory.is_full:
            num_matches = 0
            # match according to meshgrid
            for item1 in reversed_data:
                for item2 in memory:
                    num_matches += int(is_equivalent_tuple(item1, item2))

            assert (
                num_matches == bulk_size
            ), f"""Expected {bulk_size} matches inside the memory, got {num_matches}."""

            continue

        for step in range(bulk_size):
            item1 = reversed_data[step]
            item2 = memory.__getitem__((iteration * bulk_size + step) % mem_size)
            assert is_equivalent_tuple(
                item1, item2
            ), f"""Something went wrong with rollover at iteration {iteration},
                step {step}, expected {item1}, got {item2}."""


@pytest.mark.parametrize("random_rollover", [True, False])
def test_non_bulk(random_rollover):
    """Tests the replay buffer generically."""
    mem_size = 11
    element_shapes = [(3, 3), (3,), ()]
    memory = ReplayBuffer(mem_size=mem_size)

    previous_data = []
    for iteration in range(20):
        current_data = []
        for shape in element_shapes:
            current_data.append(np.random.randn(*shape))
        memory.push(current_data, random_rollover=random_rollover)

        # if random rollover and we're more than full, different matching method
        if random_rollover and memory.is_full:
            num_current_matches = 0
            num_previous_matches = 0
            for item in memory:
                num_current_matches += int(is_equivalent_tuple(item, current_data))
                num_previous_matches += int(is_equivalent_tuple(item, previous_data))

            assert (
                num_current_matches == 1
            ), f"""Expected 1 match for current_data, got {num_current_matches}."""
            assert (
                num_previous_matches <= 1
            ), f"""Expected 1 or 0 match for previous_data, got {num_previous_matches}."""

            continue

        # check the current data
        output = memory.__getitem__(iteration % mem_size)
        assert is_equivalent_tuple(
            output, current_data
        ), f"""Something went wrong with rollover at iteration {iteration},
            expected {current_data}, got {output}."""

        # check the previous data
        if iteration > 0:
            output = memory.__getitem__((iteration - 1) % mem_size)
            assert is_equivalent_tuple(
                output, previous_data
            ), f"""Something went wrong with rollover at iteration {iteration},
                expected {previous_data}, got {output}."""

        previous_data = deepcopy(current_data)



"""Some simple Wingman functionality."""
from __future__ import annotations

import numpy as np

try:
    import torch
except ImportError as e:
    raise ImportError(
        "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
    ) from e


def gpuize(input, device) -> torch.Tensor:
    """gpuize.

    Args:
        input: the array that we want to gpuize
        device: a string of the device we want to move the thing to
    """
    if torch.is_tensor(input):
        if input.device == device:
            return input.float()
        return input.to(device).float()
    return torch.tensor(input).float().to(device)


def cpuize(input) -> np.ndarray:
    """cpuize.

    Args:
        input: the array of the thing we want to put on the cpu
    """
    if torch.is_tensor(input):
        return input.detach().cpu().numpy()
    else:
        return input


def shutdown_handler(*_):
    """shutdown_handler.

    Args:
        _:
    """
    print("ctrl-c invoked")
    exit(0)



"""CLI functionality of Wingman."""
import glob
import os
import shutil
import sys

from .print_utils import cstr, wm_print


def _get_dir_size(path):
    total = 0
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                total += entry.stat().st_size
            elif entry.is_dir():
                total += _get_dir_size(entry.path)
    return total


def compress_weights():
    """Compresses the weights directory by deleting empty directories and compressing all mark numbers down to 0.

    Example CLI usage:
    `wingman-compress-weights [optional weights directory]`
    """
    wm_print("---------------------------------------------")
    wm_print("Beginning compression...")
    wm_print("---------------------------------------------")

    target_dir = "./weights"
    if len(sys.argv) > 1:
        target_dir = sys.argv[1]

    # get the original filesize
    original_size = _get_dir_size(target_dir)

    # all versions
    versions = glob.glob(f"./{target_dir}/Version*")

    # debug version, remove from all versions
    debug_version = f"./{target_dir}/VersionDebug"
    if debug_version in versions:
        versions.remove(debug_version)

    # which folders to delete
    empty_versions = []

    # start going through each folder
    for version in versions:
        marks = glob.glob(os.path.join(version, "weights*.pth"))

        # if no weights files
        if len(marks) == 0:
            empty_versions.append(version)
            continue

        # start finding the latest file
        mark_number = 0
        latest_mark = os.path.join(version, f"weights{mark_number}.pth")
        while latest_mark in marks:
            # as long as it exists, increment it
            mark_number += 1
            latest_mark = os.path.join(version, f"weights{mark_number}.pth")

        # remove the latest mark number from the list, we want to keep this
        mark_number = min(0, mark_number - 1)
        latest_mark = os.path.join(version, f"weights{mark_number}.pth")
        if latest_mark in marks:
            marks.remove(latest_mark)

        # remove the intermediary from our list if it's there
        intermediary = os.path.join(version, "weights-1.pth")
        if intermediary in marks:
            marks.remove(intermediary)

        # exit if nothing to delete
        if len(marks) == 0:
            wm_print(f"Nothing to compress in {cstr(version, 'OKGREEN')}.")
            continue

        # for the remainder of the list, delete all the weights
        wm_print(f"Compressing {version}...")
        for weight in marks:
            os.remove(weight)

        # rename the latest weights to 0
        os.rename(latest_mark, os.path.join(version, "weights0.pth"))

        wm_print(f"Compressed {cstr(version, 'WARNING')}.")

    # delete flagged directories
    for version in empty_versions:
        shutil.rmtree(version, ignore_errors=False, onerror=None)
        wm_print(f"Deleted {cstr(version, 'WARNING')}.")

    # delete the debug directory
    if debug_version in versions:
        shutil.rmtree(debug_version, ignore_errors=False, onerror=None)
        wm_print(f"Deleted {cstr(debug_version, 'WARNING')}.")

    # get the final filesize
    final_size = _get_dir_size(target_dir)

    # printout
    wm_print("---------------------------------------------")
    wm_print(f"Original disk usage: {cstr(original_size / 1e9, 'OKGREEN')} gigabytes.")
    wm_print(
        f"Removed {cstr((original_size - final_size) / 1e9, 'WARNING')} gigabytes."
    )
    wm_print(f"Final disk usage: {cstr(final_size / 1e9, 'WARNING')} gigabytes.")
    wm_print("---------------------------------------------")


def generate_yaml():
    """Generates the basic yaml file for the Wingman module.

    Example CLI usage:
    `wingman-generate-yaml [optional filename]`
    """
    source = os.path.join(os.path.dirname(__file__), "./config.yaml")

    target = "./config.yaml"
    if len(sys.argv) > 1:
        target = sys.argv[1]

    shutil.copyfile(source, target)

    wm_print(f"Generated config file at: {cstr(target, 'OKGREEN')}.")



"""The core of Wingman."""
from __future__ import annotations

import argparse
import math
import os
import time
from typing import Tuple

import numpy as np
import wandb
import yaml

from .print_utils import cstr, wm_print


class Wingman:
    """Wingman.

    Class to handle checkpointing of loss and handling of weights files.
    Minimal example:

    ```py
    helper = Wingman(./config.yaml)
    cfg = helper.cfg

    # load the model and optimizer
    model = Model().to(cfg.device)
    optim = optimizer.AdamW(model.parameters(), lr=cfg.learning_rate, amsgrad=True)

    # get the weight files if they exist
    have_file, weight_file, optim_file = self.get_weight_files()
    if have_file:
        model.load(model_file)
        optim.load(optim_file)

    # run some training:
    while(training):
        ...
        # training code here
        ...

        update_weights, model_file, optim_file = self.checkpoint(loss, step_number)
        if update_weights:
            model.save(model_file)
            optim.save(optim_file)
    ```
    """

    def __init__(
        self,
        config_yaml: str,
        experiment_description: str = "",
    ):
        """__init__.

        Args:
            config_yaml (str): location of where the config yaml is described
            experiment_description (str): optional description of the experiment
        """
        # save our experiment description
        self.config_yaml = config_yaml
        self.experiment_description = experiment_description
        self.cfg = self.__yaml_to_args()

        # make sure that logging_interval is positive
        assert (
            self.cfg.logging_interval > 0
        ), f"logging_interval must be a positive number, got {self.cfg.logging_interval}."

        # the logger
        self.log = dict()

        # runtime variables
        self.num_losses = 0
        self.cumulative_loss = 0
        self.lowest_loss = math.inf
        self.next_log_step = self.cfg.logging_interval
        self.previous_checkpoint_step = 0
        self.skips = 0

        # weight file variables
        self.directory = os.path.dirname(__file__)
        self.version_number = self.cfg.version_number
        self.mark_number = self.cfg.mark_number
        self.previous_mark_number = -1

        # directory itself
        self.version_directory = os.path.join(
            self.cfg.weights_directory, f"Version{self.version_number}"
        )

        # file paths
        self.model_file = os.path.join(
            self.version_directory,
            f"weights{self.mark_number}.pth",
        )
        self.optim_file = os.path.join(
            self.version_directory,
            "optimizer.pth",
        )
        self.status_file = os.path.join(
            self.version_directory,
            "lowest_loss.npy",
        )
        self.intermediary_file = os.path.join(
            self.version_directory,
            "weights-1.pth",
        )
        if self.log_file:
            self.log_file = os.path.join(
                self.version_directory,
                "log.txt",
            )

        wm_print("--------------𓆩𓆪--------------")
        wm_print(f"Using device {cstr(self.device, 'HEADER')}")
        wm_print(f"Saving weights to {cstr(self.version_directory, 'HEADER')}...")

        # check to record that we're in a new training session
        self.fresh_directory = False
        if not os.path.isdir(self.version_directory):
            self.fresh_directory = True
            wm_print(
                cstr(
                    "New training instance detected, generating weights directory in 3 seconds...",
                    "WARNING",
                ),
            )
            time.sleep(3)
            os.makedirs(self.version_directory)

    def __get_device(self):
        """__get_device."""
        from warnings import warn

        try:
            import torch

            if torch.cuda.is_available():
                self.device = torch.device("cuda:0")
            else:
                self.device = "cpu"
        except ImportError:
            warn(
                "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
            )
            self.device = "Not Found!"

        return self.device

    def __yaml_to_args(self):
        """__yaml_to_args.

        Reads the yaml file provided at init and converts it to commandline arguments.

        """
        # parse the arguments
        parser = argparse.ArgumentParser(description=self.experiment_description)

        with open(self.config_yaml) as f:
            # read in the file
            config = yaml.load(f, Loader=yaml.FullLoader)

            # checks that we have the default params in the file
            assertation_list = [
                "debug",
                "weights_directory",
                "version_number",
                "mark_number",
                "log_status",
                "increment",
                "logging_interval",
                "max_skips",
                "greater_than",
                "wandb",
                "wandb_name",
                "wandb_notes",
                "wandb_id",
                "wandb_entity",
                "wandb_project",
            ]
            for item in assertation_list:
                assert item in config, cstr(
                    f"Missing parameter {item} in config file.", "FAIL"
                )

            # override version number if needed
            if config["version_number"] is None:
                config["version_number"] = np.random.randint(999999)

            # make a logging file if required
            self.log_file = config["log_status"]

            # add all to argparse
            for item in config:
                parser.add_argument(
                    f"--{item}",
                    type=type(config[item]),
                    nargs="?",
                    const=True,
                    default=config[item],
                    help="None",
                )

        # dict of all arguments, by default will be overridden by commandline args
        config = {**vars(parser.parse_args())}

        # add the gpu
        config["device"] = self.__get_device()

        # change the version if debugging
        config["version_number"] = (
            "Debug" if config["debug"] else str(config["version_number"])
        )

        # cfg depending on whether wandb is enabled
        cfg = None
        if config["wandb"]:
            wandb.init(
                project=config["wandb_project"],
                entity=config["wandb_entity"],
                config=config,
                name=config["wandb_name"] + ", v=" + config["version_number"]
                if config["wandb_name"] != ""
                else config["version_number"],
                notes=config["wandb_notes"],
                id=config["wandb_id"] if config["wandb_id"] != "" else None,
            )

            # also save the code if wandb
            # wandb.run.log_code(".", exclude_fn=lambda path: "venv" in path)  # type: ignore

            # set to be consistent with wandb config
            cfg = wandb.config
        else:
            # otherwise just merge settings with args
            cfg = argparse.Namespace(**config)

        return cfg

    def checkpoint(self, loss: float, step: int | None = None) -> Tuple[bool, str, str]:
        """checkpoint.

        Records training every logging_interval steps.

        Returns three things:
        - indicator on whether we should save weights
        - directory of where the weight files should be saved
        - directory of where the optim files should be saved

        Args:
            loss (float): learning loss of the model as a detached float
            step (int | None): step number, automatically incremented if None

        Returns:
            Tuple[bool, str, str]: to_update, weights_file, optim_file
        """
        # if step is None, we automatically increment
        if step is None:
            step = self.previous_checkpoint_step + 1

        # check that our step didn't go in reverse
        assert step >= self.previous_checkpoint_step, cstr(
            f"We can't step backwards! Got step {step} but the previous logging step was {self.previous_checkpoint_step}.",
            "FAIL",
        )
        self.previous_checkpoint_step = step

        """ACCUMULATE LOSS"""
        self.cumulative_loss += loss
        self.num_losses += 1.0

        # if we haven't passed the required number of steps
        if step < self.next_log_step:
            return False, self.model_file, self.optim_file

        # log to wandb if needed, but only on the logging steps
        if self.cfg.wandb:
            self.wandb_log()

        """GET NEW AVG LOSS"""
        # record the losses and reset the cumulative
        avg_loss = self.cumulative_loss / self.num_losses
        self.cumulative_loss = 0.0
        self.num_losses = 0.0

        # compute the next time we need to log
        self.next_log_step = (
            int(step / self.cfg.logging_interval) + 1
        ) * self.cfg.logging_interval

        # always print on n intervals
        wm_print(
            f"Step {cstr(step, 'OKCYAN')}; Average Loss {cstr(f'{avg_loss:.5f}', 'OKCYAN')}; Lowest Average Loss {cstr(f'{self.lowest_loss:.5f}', 'OKCYAN')}",
            self.log_file,
        )

        """CHECK IF NEW LOSS IS BETTER"""
        # if we don't meet the criteria for saving
        if avg_loss >= self.lowest_loss - self.cfg.greater_than:
            # accumulate skips
            if self.skips < self.cfg.max_skips:
                self.skips += 1
                return False, self.model_file, self.optim_file
            else:
                # save the network to intermediary if we crossed the max number of skips
                wm_print(
                    f"Passed {self.cfg.max_skips} intervals without saving so far, saving weights to: {cstr(self.intermediary_file, 'OKCYAN')}",
                    self.log_file,
                )
                self.skips = 0
                return True, self.intermediary_file, self.optim_file

        """NEW LOSS IS BETTER"""
        # redefine the new lowest loss and reset the skips
        self.lowest_loss = avg_loss
        self.skips = 0

        # increment means return the files with incremented mark number
        if self.cfg.increment:
            # check if we are safe to increment mark numbers and regenerate weights file
            if self.previous_mark_number == -1 or os.path.isfile(self.model_file):
                self.previous_mark_number = self.mark_number
                self.model_file = os.path.join(
                    self.version_directory,
                    f"weights{self.mark_number}.pth",
                )
                self.mark_number += 1

            else:
                wm_print(
                    cstr(
                        "Didn't save weights file for the previous mark number (self.mark_number), not incrementing mark number.",
                        "WARNING",
                    ),
                    self.log_file,
                )
                self.mark_number = self.previous_mark_number

        # record the lowest running loss in the status file
        np.save(self.status_file, self.lowest_loss)

        wm_print(
            f"New lowest point, saving weights to: {cstr(self.model_file, 'OKGREEN')}",
            self.log_file,
        )

        return True, self.model_file, self.optim_file

    def wandb_log(self) -> None:
        """wandb_log.

        Logs the internal log to WandB.
        Start logging by adding things to it using:
        ```
        wm.log["foo"] = "bar"
        ```

        Returns:
            None:
        """
        assert isinstance(self.log, dict), cstr(
            f"log must be dictionary, currently it is {self.log}.", "FAIL"
        )
        if self.cfg.wandb:
            wandb.log(self.log)

    def write_auxiliary(
        self, data: np.ndarray, variable_name: str, precision: str = "%1.3f"
    ) -> None:
        """write_auxiliary.

        Args:
            data (np.ndarray): data
            variable_name (str): variable_name
            precision (str): precision

        Returns:
            None:
        """
        assert len(data.shape) == 1, "Data must be only 1 dimensional ndarray"
        filename = os.path.join(self.version_directory, f"{variable_name}.csv")
        with open(filename, "ab") as f:
            np.savetxt(f, [data], delimiter=",", fmt=precision)

    def get_weight_files(self, latest: bool = True) -> Tuple[bool, str, str]:
        """get_weight_files.

        Returns three things:
        - indicator on whether we have weight files
        - directory of where the weight files are
        - directory of where the optim files are

        Args:
            latest (bool): whether we want the latest file or the one determined by `mark_number`

        Returns:
            Tuple[bool, str, str]: have_file, weights_file, optim_file
        """
        # if we don't need the latest file, get the one specified
        if not latest:
            if os.path.isfile(self.model_file):
                self.model_file = os.path.join(
                    self.version_directory,
                    f"weights{self.mark_number}.pth",
                )
                wm_print(
                    f"Using weights file: {cstr(f'{self.version_directory}/weights{self.mark_number}.pth', 'OKGREEN')}",
                    self.log_file,
                )
                return True, self.model_file, self.optim_file
            else:
                raise ValueError(
                    cstr(
                        f"Mark number {self.mark_number} was requested, but it doesn't exist.",
                        "FAIL",
                    )
                )

        # while the file exists, try to look for a file one version later
        self.mark_number = 0
        while os.path.isfile(self.model_file):
            self.mark_number += 1
            self.model_file = os.path.join(
                self.version_directory,
                f"weights{self.mark_number}.pth",
            )

        # once the file version doesn't exist, decrement by one and use that file
        self.mark_number = max(self.mark_number - 1, 0)
        self.model_file = os.path.join(
            self.version_directory,
            f"weights{self.mark_number}.pth",
        )

        # if the file doesn't exist, notify and ignore
        if not os.path.isfile(self.model_file):
            if not self.fresh_directory:
                wm_print(
                    cstr(
                        "No weights file found, generating new one during training.",
                        "WARNING",
                    ),
                    self.log_file,
                )
            self.fresh_directory = False

            return False, self.model_file, self.optim_file
        else:
            # hitch a ride to update the lowest running loss
            self.lowest_loss = np.load(self.status_file).item()

            wm_print(
                f"Using weights file: {cstr(f'{self.version_directory}/weights{self.mark_number}.pth', 'OKGREEN')}",
                self.log_file,
            )
            wm_print(
                f"Lowest Running Loss for Net: {cstr(self.lowest_loss, 'OKCYAN')}",
                self.log_file,
            )

            # check if the optim file exists
            if not os.path.isfile(self.optim_file):
                wm_print(
                    cstr("Optim file not found, please be careful!", "WARNING"),
                    self.log_file,
                )

            return True, self.model_file, self.optim_file



"""Replay buffer implementation with push, automatic overflow, and automatic torch dataset functionality."""
from __future__ import annotations

from typing import Any, Sequence

import numpy as np

try:
    from torch.utils.data import Dataset
except ImportError as e:
    raise ImportError(
        "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
    ) from e

from .print_utils import cstr, wm_print


class ReplayBuffer(Dataset):
    """Replay Buffer implementation of a Torch dataset."""

    def __init__(self, mem_size: int):
        """__init__.

        Args:
            mem_size (int): number of transitions the replay buffer aims to hold
        """
        self.memory = []
        self.mem_size = int(mem_size)
        self.count = 0

    def __len__(self):
        """__len__."""
        return min(self.mem_size, self.count)

    def __getitem__(self, idx):
        """__getitem__.

        Args:
            idx: index of the tuple to get from the replay buffer
        """
        return list(d[idx] for d in self.memory)

    def __repr__(self):
        """Printouts parameters of this replay buffer."""
        return f"""ReplayBuffer of size {self.mem_size} with {len(self.memory)} elements. \n
        Element shapes are {[elem.shape[1:] for elem in self.memory]}. \n
        A brief view of the memory: \n
        {self.memory}
        """

    @staticmethod
    def __ensure_dims(thing: Any, bulk: bool) -> np.ndarray:
        """Ensures that all arrays are at least [n, ...] and not [n, ].

        Args:
            thing: input

        Returns:
            np.ndarray: output
        """
        thing = np.asarray(thing)
        if len(thing.shape) == int(bulk):
            thing = np.expand_dims(thing, axis=-1)
        return thing

    def push(
        self,
        data: Sequence[np.ndarray | float | int | bool],
        bulk: bool = False,
        random_rollover: bool = False,
    ):
        """Adds transition tuples into the replay buffer.

        The data must be either:
        - an n-long tuple of a single transition
        - an n-long tuple of m transitions, ie: a list of [m, ...] np arrays with the `bulk` flag set to True

        Args:
            data (Sequence[np.ndarray | float | int | bool]): data
            bulk (bool): whether to bulk add stuff into the replay buffer
            random_rollover (bool): whether to rollover the data in the replay buffer once full or to randomly insert
        """
        # check if we are bulk adding things in and assert lengths
        bulk_size = 1
        if bulk:
            assert all([isinstance(d, np.ndarray) for d in data]), cstr(
                "All things must be np.ndarray for bulk data.", "FAIL"
            )

            bulk_size = data[0].shape[0]  # pyright: ignore

            assert all([len(d) == bulk_size for d in data]), cstr(  # pyright: ignore
                f"All things in data must have same len for the first dimension for bulk data. Received data with {[len(d) for d in data]} items respectively.",  # pyright: ignore
                "FAIL",
            )

            # assert on memory lengths
            assert self.mem_size >= bulk_size, cstr(
                f"Bulk size ({bulk_size}) should be less than or equal to memory size ({self.mem_size}).",
                "FAIL",
            )

        np_data = list(map(lambda x: self.__ensure_dims(x, bulk), data))

        # instantiate the memory if it does not exist
        if self.count == 0:
            self.memory = []
            for thing in np_data:
                if not bulk:
                    self.memory.append(
                        np.zeros((self.mem_size, *thing.shape), dtype=np.float64)
                    )
                else:
                    self.memory.append(
                        np.zeros((self.mem_size, *thing.shape[1:]), dtype=np.float64)
                    )

            mem_size_bytes = sum([d.nbytes for d in self.memory])
            wm_print(
                cstr(f"Replay Buffer Size: {mem_size_bytes / 1e9} gigabytes.", "OKCYAN")
            )

        # assert that the number of lists in memory is same as data to push
        assert len(np_data) == len(self.memory), cstr(
            f"Data length not similar to memory buffer length. Replay buffer has {len(self.memory)} items, but received {len(np_data)} items.",
            "FAIL",
        )

        # indexing for memory positions
        start = self.count % self.mem_size
        stop = min(start + bulk_size, self.mem_size)
        rollover = -(self.mem_size - start - bulk_size)
        if random_rollover:
            if not self.is_full:
                idx_start = np.arange(start, stop)
                idx_stop = np.random.choice(
                    start, size=np.maximum(rollover, 0), replace=False
                )
            else:
                idx_start = np.random.choice(
                    self.mem_size, size=bulk_size, replace=False
                )
                idx_stop = []
        else:
            idx_start = np.arange(start, stop)
            idx_stop = np.arange(0, rollover)
        indices = np.append(idx_start, idx_stop).astype(np.int64)

        # put things in memory
        for memory, thing in zip(self.memory, np_data):
            memory[indices] = thing

        self.count += bulk_size

    @property
    def is_full(self) -> bool:
        """Whether or not the replay buffer has reached capacity.

        Returns:
            bool: whether the buffer is full
        """
        return self.count >= self.mem_size
